import numpy as np
import pandas as pd
from sklearn.base import TransformerMixin, BaseEstimator
from typing import Union, Optional


class SigmaClipper(TransformerMixin, BaseEstimator):
    """Clip values exceeding specified sigma levels from median
    
    :param sigma: both low_sigma and high_sigma
    :param low_sigma: low sigma level (aabsolute value)
        Overrides sigma
    :param max_sigma: high sigma level (aabsolute value)
        Overrides sigma
    :param mean_fun: 'median' | 'mean'
    """

    def __init__(self, sigma: Optional[float] = 3.,
                 low_sigma: Optional[float] = None,
                 high_sigma: Optional[float] = None,
                 mean_fun: str = 'median'):
        self.sigma = sigma
        self.low_sigma = low_sigma
        self.high_sigma = high_sigma
        self.mean_fun = mean_fun

    def fit(self, X: Union[np.array, pd.DataFrame], y=None):
        self.low_sigma_ = self.sigma if self.low_sigma is None else self.low_sigma
        self.high_sigma_ = self.sigma if self.high_sigma is None else self.high_sigma
        assert self.low_sigma_ >= 0
        assert self.high_sigma_ >= 0
        
        if self.mean_fun == 'median':
            meanf = np.nanmedian
        elif self.mean_fun == 'mean':
            meanf = np.nanmean
        else:
            raise ValueError(self.mean_fun)
        X_ = np.asarray(X)
        self.mean_ = meanf(X_, axis=0)
        self.std_ = np.sqrt(meanf((X_ - self.mean_) ** 2, axis=0))
        self.high_ = self.mean_ + self.low_sigma_ * self.std_
        self.low_ = self.mean_ - self.high_sigma_ * self.std_
        return self
        
    def transform(self, X: Union[np.array, pd.DataFrame]) -> Union[np.array, pd.DataFrame]:
        if isinstance(X, pd.DataFrame):
            X = X.clip(self.low_, self.high_, axis=1)
        else:
            X = np.clip(X, self.low_[np.newaxis, :], self.high_[np.newaxis, :])
        return X


class QuantileClipper(TransformerMixin, BaseEstimator):
    """Clip values exceeding specified quantile levels times a factor
    
    :param factor: multiplier
    :param low_quantile: low quantile level
    :param high_quantile: high quantile level
    :param mean_fun: 'median' | 'mean'
    """

    def __init__(self, factor: Optional[float] = 3.,
                 low_quantile: Optional[float] = 0.9,
                 high_quantile: Optional[float] = 0.1,
                 mean_fun: str = 'median'):
        self.factor = factor
        self.low_quantile = low_quantile
        self.high_quantile = high_quantile
        self.mean_fun = mean_fun

    def fit(self, X: Union[np.array, pd.DataFrame], y=None):
        assert self.factor >= 0
        X_ = np.asarray(X)
        self.mean_ = np.nanmedian(X_, axis=0)
        self.high_q_ = np.nanquantile(X_, self.high_quantile, axis=0)
        self.low_q_ = np.nanquantile(X_, self.low_quantile, axis=0)
        self.high_ = (self.high_q_ - self.mean_) * self.factor + self.mean_
        self.low_ = (self.low_q_ - self.mean_) * self.factor + self.mean_
        return self
        
    def transform(self, X: Union[np.array, pd.DataFrame]) -> Union[np.array, pd.DataFrame]:
        if isinstance(X, pd.DataFrame):
            X = X.clip(self.low_, self.high_, axis=1)
        else:
            X = np.clip(X, self.low_[np.newaxis, :], self.high_[np.newaxis, :])
        return X



import pandas as pd
from sklearn.base import TransformerMixin, BaseEstimator
from typing import Optional, List


class ColumnSelector(TransformerMixin, BaseEstimator):
    """Transformer affecting only selected columns"""

    def __init__(self, transformer=None,
                 columns: Optional[List[str]] = None,
                 remainder: str = 'passthrough',
                 copy: bool = True):
        self.transformer = transformer
        self.columns = columns
        self.remainder = remainder
        self.copy = copy

    def fit(self, X: pd.DataFrame, y=None):
        if self.transform is None:
            return self
        self.columns_ = self.columns if self.columns is not None else X.columns
        self.transformer.fit(X=X[self.columns_], y=y)
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.transform is None:
            return X
        if self.remainder != 'passthrough':
            raise NotImplementedError
        if self.copy:
            X = X.copy()
        X_transformed = self.transformer.transform(X=X[self.columns_])
        if isinstance(X_transformed, pd.DataFrame):
            for c in self.columns_:
                X[c] = X_transformed[c]
        else:
            X.loc[:, self.columns_] = X_transformed
        return X

    def inverse_transform(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.transform is None:
            return X
        if hasattr(self.transformer, 'inverse_transform'):
            if self.copy:
                X = X.copy()
            X_orig = self.transformer.inverse_transform(X=X[self.columns_])
            if isinstance(X_orig, pd.DataFrame):
                for c in self.columns_:
                    X[c] = X_orig[c]
            else:
                X.loc[:, self.columns_] = X_orig
            return X
        else:
            return None



import numpy as np
from sklearn.model_selection import cross_validate
from sklearn.linear_model import LinearRegression
from mlutil.eval import TimeSeriesSplit


def test_dummy():
    assert True


def test_TimeSeriesSplit():
    X = np.vstack([np.random.normal(size=100), np.random.normal(size=100)]).T
    y = np.random.normal(size=100)
    m = LinearRegression()
    cv = TimeSeriesSplit(test_period_len=50)
    scores = cross_validate(m, X, y, scoring=['neg_mean_squared_error'], cv=cv)
    assert len(scores['test_neg_mean_squared_error']) > 0

    cv2 = TimeSeriesSplit(test_period_len=0.5)
    scores2 = cross_validate(m, X, y, scoring=['neg_mean_squared_error'], cv=cv2)
    np.testing.assert_allclose(scores['test_neg_mean_squared_error'], scores2['test_neg_mean_squared_error'])



import pytest
import numpy as np
import pandas as pd
from mlutil.transform import SigmaClipper, QuantileClipper, ColumnSelector


@pytest.mark.parametrize(
    'X, X_new, sigma, sigma_low, sigma_high',
    [
        (
            pd.DataFrame({
                'a': [np.nan, -1., 2., 1., 1., 302.],
                'b': [-2., 1., 3., 2., -201, np.nan],
            }),
            pd.DataFrame({
                'a': [np.nan, -1., 2., 1., 1., 4.],
                'b': [-2., 1., 3., 2., -5, np.nan],
            }),
            3., None, None,
        ),
        (
            pd.DataFrame({
                'a': [np.nan, -1., 2., 1., 1., 302.],
                'b': [-2., 1., 3., 2., -201, np.nan],
            }),
            pd.DataFrame({
                'a': [np.nan, -1., 2., 1., 1., 4.],
                'b': [-2., 1., 3., 2., -5, np.nan],
            }),
            1., 3., 3.,
        ),
        (
            np.array([
                [np.nan, -1., 2., 1., 1., 302.],
                [-2., 1., 3., 2., -201, np.nan],
            ]).T,
            np.array([
                [np.nan, -1., 2., 1., 1., 4.],
                [-2., 1., 3., 2., -5, np.nan],
            ]).T,
            3., None, None,
        ),
    ]
)
def test_SigmaClipper(X, X_new, sigma, sigma_low, sigma_high):
    t = SigmaClipper(sigma=sigma, low_sigma=sigma_low, high_sigma=sigma_high)
    X_new_ = t.fit_transform(X)
    if isinstance(X, np.ndarray):
        np.testing.assert_allclose(X_new_, X_new)
    elif isinstance(X, pd.DataFrame):
        np.testing.assert_allclose(X_new_.values, X_new.values)
    else:
        raise TypeError(type(X))


@pytest.mark.parametrize(
    'X, X_new, factor, q_low, q_high',
    [
        (
            pd.DataFrame({
                'a': [np.nan, -1., 2., 0, 1., 302.],
                'b': [-2., 1., 3., 2., -201, np.nan],
            }),
            pd.DataFrame({
                'a': [np.nan, -1., 2., 0, 1., 4.],
                'b': [-2., 1., 3., 2., -8., np.nan],
            }),
            3., .25, .75,
        ),
        (
            np.array([
                [np.nan, -1., 2., 0, 1., 302.],
                [-2., 1., 3., 2., -201, np.nan],
            ]).T,
            np.array([
                [np.nan, -1., 2., 0, 1., 4.],
                [-2., 1., 3., 2., -8., np.nan],
            ]).T,
            3., .25, .75,
        ),
    ]
)
def test_QuantileClipper(X, X_new, factor, q_low, q_high):
    t = QuantileClipper(factor=factor, low_quantile=q_low, high_quantile=q_high)
    X_new_ = t.fit_transform(X)
    if isinstance(X, np.ndarray):
        np.testing.assert_allclose(X_new_, X_new)
    elif isinstance(X, pd.DataFrame):
        np.testing.assert_allclose(X_new_.values, X_new.values)
    else:
        raise TypeError(type(X))


@pytest.mark.parametrize(
    'X, X_new, columns',
    [
        (
            pd.DataFrame({
                'a': [np.nan, -1., 2., 1., 1., 302.],
                'b': [-2., 1., 3., 2., -201, np.nan],
            }),
            pd.DataFrame({
                'a': [np.nan, -1., 2., 1., 1., 4.],
                'b': [-2., 1., 3., 2., -5, np.nan],
            }),
            None,
        ),
        (
            pd.DataFrame({
                'a': [np.nan, -1., 2., 1., 1., 302.],
                'b': [-2., 1., 3., 2., -201, np.nan],
            }),
            pd.DataFrame({
                'a': [np.nan, -1., 2., 1., 1., 4.],
                'b': [-2., 1., 3., 2., -201, np.nan],
            }),
            ['a'],
        ),
    ]
)
def test_ColumnSelector(X, X_new, columns):
    t = ColumnSelector(
        SigmaClipper(low_sigma=3, high_sigma=3),
        columns=columns,
    )
    X_new_ = t.fit_transform(X)
    if isinstance(X, np.ndarray):
        np.testing.assert_allclose(X_new_, X_new)
    elif isinstance(X, pd.DataFrame):
        np.testing.assert_allclose(X_new_.values, X_new.values)
    else:
        raise TypeError(type(X))



import numpy as np
from statsmodels.gam.api import GLMGam, BSplines, CyclicCubicSplines
from sklearn.base import RegressorMixin, BaseEstimator
from typing import Union, List


class GAM(RegressorMixin, BaseEstimator):
    """Generalized Additive Models
    
    :param splines: type of sline functions
        'cyclic_cubic' | 'b'
    :param df: number of spline basis functions (= knots)
        if splines='cyclic_cubic', must be set to 3 (default value)
    :param degree: degree of splines
    :param family: distribution gamily (default: gaussian)
        See statsmodels.api.families
    :param alpha: penalization weight
    :param clip_X: if true, clip X values by its min and max during training
        use for 'b' splines, which do not have extrapolation
    """

    def __init__(
        self,
        splines: str = 'cyclic_cubic',
        df: Union[int, List[int]] = 5,
        degree: Union[int, List[int]] = 3,
        family=None,
        alpha: Union[float, List[float]] = 0.01,
        clip_X: bool = False,
    ):
        self.df = df
        self.degree = degree
        self.family = family
        self.alpha = alpha
        self.splines = splines
        self.clip_X = clip_X
    
    def fit(self, X, y):
        X = np.array(X)
        assert not (self.splines == 'cyclic_cubic') or (self.degree == 3)
        df = self.df if isinstance(self.df, list) else [self.df] * X.shape[1]
        degree = self.degree if isinstance(self.degree, list) else [self.degree] * X.shape[1]
        alpha = self.alpha if isinstance(self.alpha, list) else [self.alpha] * X.shape[1]
        if self.splines == 'cyclic_cubic':
            self.splines_ = CyclicCubicSplines(X, df=df)
        elif self.splines == 'b':
            self.splines_ = BSplines(X, df=df, degree=degree)
        else:
            raise ValueError(self.splines)
        self.x_min_ = np.min(X, axis=0)
        self.x_max_ = np.max(X, axis=0)
        self.estimator_ = GLMGam(y, X, smoother=self.splines_, family=self.family, alpha=alpha)
        self.res_ = self.estimator_.fit()
        return self
    
    def predict(self, X):
        if self.clip_X:
            X = np.clip(X, self.x_min_, self.x_max_)
        return self.res_.predict(X, exog_smooth=X)



from typing import Optional, Union


class TimeSeriesSplit:
    """Split data for nested (rolling) time-series cross-validation"""

    def __init__(self,
                 test_period_len: Union[int, float],
                 train_period_len: Optional[Union[int, float]] = None,
                 n_predictions: Union[int, float] = 1,
                 prediction_lag: Union[int, float] = 0,
                 prediction_step: Union[int, float] = 1,
                 leave_n_first_points: Union[int, float] = 0,
                 leave_n_last_points: Union[int, float] = 0):
        """
        :param test_period_len: Total number of points in the test period
        :param train_period_len: Maximum number of points in the train period
            *None* means all points will be used
        :param n_predictions: Predict <x> points ahead
            E.g., *1* means that 1 value will be forecasted in every CV fold
        :param prediction_step: Place CV folds every <x> points
            A value greater than 1 will help to speed up CV
        :param prediction_lag: Lag predictions by <x> points
            A value greater than 0 means that there is a gap between the train and test period
            Must be zero for autoregressive models
        """

        assert test_period_len > 0
        assert (train_period_len is None) or (test_period_len > 0)
        assert n_predictions > 0
        assert prediction_lag >= 0
        assert leave_n_first_points >= 0
        assert leave_n_last_points >= 0

        self.test_period_len = test_period_len
        self.train_period_len = train_period_len
        self.n_predictions = n_predictions
        self.prediction_step = prediction_step
        self.prediction_lag = prediction_lag
        self.leave_n_first_points = leave_n_first_points
        self.leave_n_last_points = leave_n_last_points

    @staticmethod
    def _cv_gen(last_idx: int,
                test_period_len: int,
                train_period_len: Optional[int],
                n_predictions: int,
                prediction_lag: int,
                prediction_step: int,
                leave_n_first_points: int,
                leave_n_last_points: int):
        last_end_test_idx = last_idx - leave_n_last_points
        last_end_train_idx = last_end_test_idx - n_predictions - prediction_lag
        first_end_train_idx = last_end_train_idx - test_period_len + prediction_step
        first_end_train_idx = first_end_train_idx + (last_end_train_idx - first_end_train_idx) % prediction_step
        if first_end_train_idx < leave_n_first_points:
            raise AssertionError(f"First training interval is empty: "
                                 f"{first_end_train_idx} < {leave_n_first_points}")

        for end_train_idx in range(first_end_train_idx, last_end_train_idx + 1,
                                   prediction_step):
            if train_period_len is not None:
                start_train_idx = max(end_train_idx - train_period_len + 1, leave_n_first_points)
            else:
                start_train_idx = leave_n_first_points
            train_ids = slice(start_train_idx,
                              min(end_train_idx, last_idx) + 1)
            test_ids = slice(end_train_idx + prediction_lag + 1,
                             min(end_train_idx + prediction_lag + n_predictions, last_idx) + 1)
            yield train_ids, test_ids

    @staticmethod
    def _to_points(x, n_points):
        if x is None:
            return x
        elif isinstance(x, int):
            assert x < n_points
            return x
        else:
            assert x < 1.
            return int(x * n_points)

    def split(self, X, y=None, groups=None):
        n_points = len(X)
        return self._cv_gen(
            last_idx=n_points - 1,
            test_period_len=self._to_points(self.test_period_len, n_points),
            train_period_len=self._to_points(self.train_period_len, n_points),
            n_predictions=self._to_points(self.n_predictions, n_points),
            prediction_step=self._to_points(self.prediction_step, n_points),
            prediction_lag=self._to_points(self.prediction_lag, n_points),
            leave_n_first_points=self._to_points(self.leave_n_first_points, n_points),
            leave_n_last_points=self._to_points(self.leave_n_last_points, n_points),
        )
 
    def get_n_splits(self, X, y=None, groups=None):
        n_splits = self.test_period_len // self.prediction_step
        return n_splits



import numpy as np
from mlutil.model import GAM


def test_GAM():
    m = GAM()
    X = np.arange(20)[:, None]
    y = np.arange(20) + np.random.normal(scale=0.1, size=20)
    m.fit(X, y)
    X_test = np.arange(15, 25)[:, None]
    y_test = np.arange(15, 25)
    y_hat = m.predict(X_test)
    np.testing.assert_allclose(y_test, y_hat, atol=1.)
